import logging
from pathlib import Path
from typing import Dict, Generator, List, Optional, Tuple

from bridgebots import (
    BiddingSuit,
    BoardRecord,
    Card,
    Deal,
    DealRecord,
    Direction,
    build_lin_url,
    parse_multi_lin,
    parse_pbn,
)
from bridgebots.deal_utils import calculate_shape, count_hcp


def _calculate_trump_data(deal: Deal, board_record: BoardRecord) -> Tuple[Optional[int], Optional[int]]:
    if board_record.contract.suit == BiddingSuit.NO_TRUMP:
        return None, None
    trump_suit = board_record.contract.suit.to_suit()
    offense_trumps = []
    offense_trumps.extend(deal.player_cards[board_record.declarer])
    offense_trumps.extend(deal.player_cards[board_record.declarer.partner()])
    offense_trumps = [c for c in offense_trumps if c.suit == trump_suit]
    return len(offense_trumps), count_hcp(offense_trumps)


def _extract_board_data(deal: Deal, board_record: BoardRecord, results_path: Path) -> Dict:
    """
    :return: A dictionary of csv keys to data values for use in a csv.DictWriter
    """
    board_dict = {}
    board_dict["board_id"] = board_record.board_name
    board_dict["file"] = results_path.name
    board_dict["deal_hash"] = hash(deal)
    board_dict.update({direction.name.lower(): board_record.names[direction] for direction in Direction})
    board_dict["dealer"] = deal.dealer.name.lower()
    board_dict["vulnerable"] = _calculate_vulnerable(deal)
    board_dict["bidding"] = _create_bidding_entry(board_record)

    opener_seat, opening_bid, opener_hcp, opener_shape_str = _calculate_opener_data(deal, board_record)
    board_dict["opener"] = opener_seat
    board_dict["opening"] = opening_bid
    board_dict["opener_hcp"] = opener_hcp
    board_dict["opener_shape"] = opener_shape_str

    overcall_type, overcall, overcaller_hcp, overcaller_shape_str, contested, competitive = _calculate_overcaller_data(
        deal, board_record
    )
    board_dict["overcall_type"] = overcall_type
    board_dict["overcall"] = overcall
    board_dict["overcaller_hcp"] = overcaller_hcp
    board_dict["overcaller_shape"] = overcaller_shape_str
    board_dict["contested"] = contested
    board_dict["competitive"] = competitive

    board_dict["declarer"] = board_record.declarer.name.lower()
    board_dict["contract"] = str(board_record.contract)
    board_dict["lead"] = board_record.play_record[0] if len(board_record.play_record) > 0 else None
    board_dict["result"] = _create_contract_result(board_record)
    board_dict["tricks"] = board_record.tricks
    board_dict["score_ns"], board_dict["score_ew"] = _calculate_direction_scores(board_record)
    board_dict["score_declarer"], board_dict["score_defender"] = board_record.score, board_record.score * -1
    board_dict["link"] = build_lin_url(deal, board_record)

    board_dict["declarer_shape"] = _calculate_shape_str(deal.player_cards[board_record.declarer])
    board_dict["dummy_shape"] = _calculate_shape_str(deal.player_cards[board_record.declarer.partner()])
    board_dict["lho_shape"] = _calculate_shape_str(deal.player_cards[board_record.declarer.offset(1)])
    board_dict["rho_shape"] = _calculate_shape_str(deal.player_cards[board_record.declarer.offset(3)])

    board_dict["declarer_hcp"] = count_hcp(deal.player_cards[board_record.declarer])
    board_dict["dummy_hcp"] = count_hcp(deal.player_cards[board_record.declarer.partner()])
    board_dict["lho_hcp"] = count_hcp(deal.player_cards[board_record.declarer.offset(1)])
    board_dict["rho_hcp"] = count_hcp(deal.player_cards[board_record.declarer.offset(3)])

    trump_fit, trump_hcp = _calculate_trump_data(deal, board_record)
    board_dict["trump_fit"] = trump_fit
    board_dict["trump_hcp"] = trump_hcp

    return board_dict


def _calculate_shape_str(cards: List[Card]) -> str:
    """Calculate a shape tuple like (5,3,3,2) and convert it to a shape_str like 5332"""
    return "".join(str(s) for s in calculate_shape(cards))


def _calculate_opener_data(
    deal: Deal, board_record: BoardRecord
) -> Tuple[Optional[int], Optional[str], Optional[int], Optional[str]]:
    """
    :return: the seat which opened, the opening bid, opener's high card points, and opener's shape
    """
    opener = None
    seat = None
    opening_bid = None
    for i in range(4):
        if board_record.bidding_record[i] != "PASS":
            opener = deal.dealer.offset(i)
            seat = i + 1
            opening_bid = board_record.bidding_record[i]
            break
    if opener is None:
        return None, None, None, None
    opener_cards = deal.player_cards[opener]
    opener_hcp = count_hcp(opener_cards)
    opener_shape_str = _calculate_shape_str(opener_cards)
    return seat, opening_bid, opener_hcp, opener_shape_str


def _calculate_overcaller_data(
    deal, board_record
) -> Tuple[Optional[str], Optional[str], Optional[int], Optional[str], Optional[bool], Optional[bool]]:
    """
    :return: the overcall type (direct, balance, or sandwich) the overcall bid, overcallers's high card points,
    overcaller's shape, and a contested boolean. The boolean will be true if someone overcalled or if one of players on
    the non-opening team eventually made any bid other than PASS.
    """
    opener = None
    open_index = None
    for i in range(4):
        if board_record.bidding_record[i] != "PASS":
            opener = deal.dealer.offset(i)
            open_index = i
            break
    if opener is None:
        return None, None, None, None, None, None
    overcall_type = None
    overcaller = None
    overcall = None
    if board_record.bidding_record[open_index + 1] != "PASS":
        overcall_type = "direct"
        overcaller = opener.offset(1)
        overcall = board_record.bidding_record[open_index + 1]
    elif board_record.bidding_record[open_index + 3] != "PASS":
        overcaller = opener.offset(3)
        overcall = board_record.bidding_record[open_index + 3]
        if board_record.bidding_record[open_index + 2] == "PASS":
            overcall_type = "balance"
        else:
            overcall_type = "sandwich"

    overcaller_hcp = None
    overcaller_shape_str = None
    if overcaller:
        overcaller_cards = deal.player_cards[overcaller]
        overcaller_hcp = count_hcp(overcaller_cards)
        overcaller_shape_str = _calculate_shape_str(overcaller_cards)

    contested = False
    for i, bid in enumerate(board_record.bidding_record[open_index:]):
        if i % 2 > 0 and bid != "PASS":
            contested = True
            break

    bid_counter = [0, 0]
    for i, bid in enumerate(board_record.bidding_record):
        if bid != "PASS":
            bid_counter[i%2] += 1
    competitive = bid_counter[0] >= 2 and bid_counter[1] >= 2
    return overcall_type, overcall, overcaller_hcp, overcaller_shape_str, contested, competitive


def _calculate_vulnerable(deal: Deal) -> str:
    """
    :return: which seats were vulnerable (1-indexed)
    """
    dealer_vuln = deal.is_vulnerable(deal.dealer)
    opp_vuln = deal.is_vulnerable(deal.dealer.next())
    if dealer_vuln and opp_vuln:
        return "all"
    elif dealer_vuln:
        return "1-3"
    elif opp_vuln:
        return "2-4"
    else:
        return "none"


def _create_bidding_entry(board_record: BoardRecord) -> str:
    """
    :return: A string representing the bids, the alerts, and the announcements
    """
    bidding_copy = board_record.bidding_record.copy()
    for bid_metadata in board_record.bidding_metadata:
        if bid_metadata.alerted:
            bidding_copy[bid_metadata.bid_index] += "!"
        if bid_metadata.explanation:
            bidding_copy[bid_metadata.bid_index] += f"({bid_metadata.explanation})"
    return "|".join(bidding_copy)


def _create_contract_result(board_record: BoardRecord) -> str:
    if board_record.contract.level == 0:
        return "PASS"
    trick_delta = board_record.tricks - (board_record.contract.level + 6)
    if trick_delta > 0:
        trick_delta_str = f"+{trick_delta}"
    elif trick_delta == 0:
        trick_delta_str = "="
    else:
        trick_delta_str = str(trick_delta)
    return str(board_record.contract) + trick_delta_str


def _calculate_direction_scores(board_record: BoardRecord) -> Tuple[int, int]:
    """
    :return: A tuple of north-south and east-west scores. Useful to have both for team comparisons.
    """
    ns_score = (
        board_record.score if board_record.declarer in [Direction.NORTH, Direction.SOUTH] else -1 * board_record.score
    )
    return ns_score, ns_score * -1


def _generate_input_file_paths(input_format: str, input_path: Path) -> Generator[Path, None, None]:
    results_file_paths = []
    if input_path.is_file():
        results_file_paths.append(input_path)
    elif input_path.is_dir():
        results_file_paths.extend(input_path.rglob(f"*.{input_format}"))
    logging.debug(f"Found {len(results_file_paths)} {input_format} files to read.")
    yield from results_file_paths


def _parse_results_file(results_file_path: Path, input_format: str) -> List[DealRecord]:
    logging.debug(f"Processing {results_file_path}")
    if input_format == "lin":
        deal_records = parse_multi_lin(results_file_path)
    else:
        deal_records = parse_pbn(results_file_path)
    logging.debug(f"Found {len(deal_records)} deals in {results_file_path}")
    return deal_records



import dataclasses
import logging
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Tuple

import click

from bridgebots import BoardRecord, Deal, Direction
from bridgebots_tools.csv_utilities import _write_results
from bridgebots_tools.data_extractors import _extract_board_data, _generate_input_file_paths, _parse_results_file
from bridgebots_tools.tools_logging import configure_logging


@dataclasses.dataclass
class TraceableRecord:
    deal: Deal
    board_record: BoardRecord
    file_path: Path


def _get_direction_key(direction_comparison_type: str, board_record: BoardRecord):
    if direction_comparison_type == "any":
        return "any"
    if board_record.declarer in [Direction.NORTH, Direction.SOUTH]:
        return "ns"
    return "ew"


def _build_comparisons(
    deal_comparison_type: str,
    direction_comparison_type: str,
    left_boards: Dict[int, Dict[str, List[TraceableRecord]]],
    right_boards: Dict[int, Dict[str, List[TraceableRecord]]],
) -> List[Tuple[List[TraceableRecord], List[TraceableRecord]]]:
    comparisons = []
    for deal_hash, left_direction_dict in left_boards.items():
        if deal_hash not in right_boards:
            for direction_key, boards in left_direction_dict.items():
                comparisons.append((boards, []))
        else:
            right_direction_dict = right_boards[deal_hash]
            for right_direction_key, right_results in right_direction_dict.items():
                for left_direction_key, left_results in left_direction_dict.items():
                    if direction_comparison_type == "opposite_direction":
                        if left_direction_key != right_direction_key:
                            comparisons.append((left_results, right_results))
                    elif left_direction_key == right_direction_key:
                        comparisons.append((left_results, right_results))

    for deal_hash, right_direction_dict in right_boards.items():
        if deal_hash not in left_boards:
            for direction_key, boards in right_direction_dict.items():
                comparisons.append(([], boards))

    if deal_comparison_type == "same_deal":
        comparisons = [c for c in comparisons if len(c[0]) > 0 and len(c[1]) > 0]
    return comparisons


@click.command()
@click.option("--input_format", type=click.Choice(["lin", "pbn"], case_sensitive=False), default="lin")
@click.option(
    "--deal_comparison_type", type=click.Choice(["same_deal", "any"], case_sensitive=False), default="same_deal"
)
@click.option(
    "--direction_comparison_type",
    type=click.Choice(["same_direction", "opposite_direction", "any"], case_sensitive=False),
    default="same_direction",
)
@click.option("--lax_doubles/--strict_doubles", default=True)
@click.option("--verbose", "log_level", "-v", flag_value="verbose")
@click.option("--info", "log_level", flag_value="info", default=True)
@click.option("--quiet", "log_level", "-q", flag_value="quiet")
@click.argument("contract_sets", nargs=2)
@click.argument("input_path", type=click.Path(exists=True, file_okay=True, dir_okay=True, path_type=Path))
@click.argument(
    "output_path", type=click.Path(exists=False, file_okay=True, dir_okay=False, path_type=Path), required=False
)
def report(
    input_format: str,
    deal_comparison_type: str,
    direction_comparison_type: str,
    lax_doubles: bool,
    log_level: str,
    contract_sets: List[str],
    input_path: Path,
    output_path: Path,
):
    configure_logging(log_level)
    results_file_paths = _generate_input_file_paths(input_format, input_path)
    left_contracts, right_contracts = [contract_set_str.split(",") for contract_set_str in contract_sets]
    left_boards, right_boards = defaultdict(lambda: defaultdict(list)), defaultdict(lambda: defaultdict(list))
    for results_file_path in results_file_paths:
        deal_records = _parse_results_file(results_file_path, input_format)
        for deal_record in deal_records:
            for board_record in deal_record.board_records:
                deal_hash = hash(deal_record.deal)
                direction_key = _get_direction_key(direction_comparison_type, board_record)
                record = TraceableRecord(deal_record.deal, board_record, results_file_path)
                contract_str = str(board_record.contract).strip("X") if lax_doubles else str(board_record.contract)
                if contract_str in left_contracts:
                    left_boards[deal_hash][direction_key].append(record)
                elif contract_str in right_contracts:
                    right_boards[deal_hash][direction_key].append(record)
    compared_deals = _build_comparisons(deal_comparison_type, direction_comparison_type, left_boards, right_boards)
    logging.debug(f"Found {len(compared_deals)} comparisons")

    csv_dicts = []
    for left_results, right_results in compared_deals:
        csv_dicts.extend([_extract_board_data(tr.deal, tr.board_record, tr.file_path) for tr in left_results])
        csv_dicts.extend([_extract_board_data(tr.deal, tr.board_record, tr.file_path) for tr in right_results])

    _write_results(output_path, csv_dicts)


if __name__ == "__main__":
    report()



import sys
from pathlib import Path

import click

from bridgebots import build_lin_str
from bridgebots_tools.csv_report import _extract_team_dicts
from bridgebots_tools.csv_utilities import _get_headers, _write_results
from bridgebots_tools.data_extractors import _parse_results_file
from bridgebots_tools.tools_logging import configure_logging

_LINKS_CSV_HEADERS = ("board_id", "file", "o_link", "c_link")


@click.command()
@click.option("--input_format", type=click.Choice(["lin", "pbn"], case_sensitive=False), default="lin")
@click.option("--verbose", "log_level", "-v", flag_value="verbose")
@click.option("--info", "log_level", flag_value="info", default=True)
@click.option("--quiet", "log_level", "-q", flag_value="quiet")
@click.argument("input_dir", type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path))
@click.argument("file_name", type=str, required=True)
@click.argument("board", type=str, required=True)
@click.argument(
    "output_lin", type=click.Path(exists=False, file_okay=True, dir_okay=False, path_type=Path), required=False
)
@click.argument(
    "output_csv", type=click.Path(exists=False, file_okay=True, dir_okay=False, path_type=Path), required=False
)
def report(
    input_format: str,
    log_level: str,
    input_dir: Path,
    file_name: str,
    board: str,
    output_lin: Path,
    output_csv: Path,
):
    configure_logging(log_level)
    input_path = next(input_dir.rglob(file_name))

    deal_records = _parse_results_file(input_path, input_format)
    csv_dicts = []
    for deal_record in deal_records:
        if "o" + board in [board_record.board_name for board_record in deal_record.board_records]:
            # Write one instance of the board to the LIN file (or stdout)
            board_record = deal_record.board_records[0]
            lin_output = build_lin_str(deal_record.deal, board_record, lin_type="multi")
            output_handle = open(output_lin, "a") if output_lin else sys.stdout
            try:
                output_handle.write(lin_output + "\n")
            finally:
                if output_lin:
                    output_handle.close()
            # Now collect CSV information from the open and closed rooms
            open_dict, closed_dict = _extract_team_dicts(deal_record, input_path)
            prefix_dict = {f"o_{key}": value for key, value in open_dict.items()}
            prefix_dict.update({f"c_{key}": value for key, value in closed_dict.items()})
            csv_dicts.append(prefix_dict)

    # Write CSV to file (or stdout)
    _write_results(output_csv, csv_dicts, _get_headers(output_format="team"))


if __name__ == "__main__":
    report()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bridgebots_tools']

package_data = \
{'': ['*']}

install_requires = \
['bridgebots==0.0.11a1', 'click>=8.0.3,<9.0.0']

setup_kwargs = {
    'name': 'bridgebots-tools',
    'version': '0.0.4',
    'description': 'Tools which utilize the Bridgebots package to process data',
    'long_description': None,
    'author': 'Forrest Rice',
    'author_email': 'forrest.d.rice@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# This list can be reordered or items can be removed to modify the output columns
import csv
import logging
import sys
from pathlib import Path
from typing import Dict, List

_CSV_HEADERS = (
    "board_id",
    "file",
    "deal_hash",
    "north",
    "south",
    "east",
    "west",
    "dealer",
    "vulnerable",
    "bidding",
    "opener",
    "opening",
    "opener_hcp",
    "opener_shape",
    "overcall_type",
    "overcall",
    "overcaller_hcp",
    "overcaller_shape",
    "contested",
    "competitive",
    "declarer",
    "contract",
    "lead",
    "result",
    "tricks",
    "score_ns",
    "score_ew",
    "score_declarer",
    "score_defender",
    "link",
    "declarer_shape",
    "dummy_shape",
    "lho_shape",
    "rho_shape",
    "declarer_hcp",
    "dummy_hcp",
    "lho_hcp",
    "rho_hcp",
    "trump_fit",
    "trump_hcp")

_PREFIX_CSV_HEADERS = [f"o_{header}" for header in _CSV_HEADERS] + [f"c_{header}" for header in _CSV_HEADERS]


def _get_headers(output_format: str):
    return _PREFIX_CSV_HEADERS if output_format == "team" else _CSV_HEADERS


def _write_results(csv_path: Path, csv_dicts: List[Dict], headers: List[str] = _CSV_HEADERS):
    """
    Write csv_dicts to csv_path using the supplied headers
    """
    logging.debug(f"Writing {len(csv_dicts)} deals to {csv_path}")
    output_handle = open(csv_path, "a") if csv_path else sys.stdout
    try:
        writer = csv.DictWriter(output_handle, fieldnames=headers, extrasaction="ignore")
        writer.writeheader()
        for csv_dict in csv_dicts:
            writer.writerow(csv_dict)
    finally:
        if csv_path:
            output_handle.close()



import logging
from pathlib import Path
from typing import Dict, Tuple

import click

from bridgebots import DealRecord
from bridgebots_tools.csv_utilities import _get_headers, _write_results
from bridgebots_tools.data_extractors import (
    _extract_board_data,
    _generate_input_file_paths,
    _parse_results_file,
)
from bridgebots_tools.tools_logging import configure_logging


def _extract_team_dicts(deal_record: DealRecord, results_path: Path) -> Tuple[Dict, Dict]:
    """
    :param deal_record: a deal record with exactly two BoardRecords
    :param results_path: the path to the original lin/pbn file
    :return: A pair of open/closed room dictionaries for csv writing
    """
    if len(deal_record.board_records) != 2:
        raise ValueError(f"Invalid number of board records for a teams match: {len(deal_record.board_records)}")
    open_board, closed_board = deal_record.board_records[0], deal_record.board_records[1]
    return _extract_board_data(deal_record.deal, open_board, results_path), _extract_board_data(
        deal_record.deal, closed_board, results_path
    )


@click.command()
@click.option("--input_format", type=click.Choice(["lin", "pbn"], case_sensitive=False), default="lin")
@click.option("--output_format", type=click.Choice(["team", "individual"], case_sensitive=False), default="team")
@click.option("--verbose", "log_level", "-v", flag_value="verbose")
@click.option("--info", "log_level", flag_value="info", default=True)
@click.option("--quiet", "log_level", "-q", flag_value="quiet")
@click.argument("input_path", type=click.Path(exists=True, file_okay=True, dir_okay=True, path_type=Path))
@click.argument(
    "output_path", type=click.Path(exists=False, file_okay=True, dir_okay=False, path_type=Path), required=False
)
def report(input_format: str, output_format: str, log_level: str, input_path: Path, output_path: Path):
    configure_logging(log_level)
    results_file_paths = _generate_input_file_paths(input_format, input_path)
    csv_dicts = []
    for results_file_path in results_file_paths:
        deal_records = _parse_results_file(results_file_path, input_format)
        for deal_record in deal_records:
            try:
                if output_format == "team":
                    open_dict, closed_dict = _extract_team_dicts(deal_record, results_file_path)
                    prefix_dict = {f"o_{key}": value for key, value in open_dict.items()}
                    prefix_dict.update({f"c_{key}": value for key, value in closed_dict.items()})
                    csv_dicts.append(prefix_dict)
                else:
                    csv_dicts.extend(
                        [
                            _extract_board_data(deal_record.deal, board_record, results_file_path)
                            for board_record in deal_record.board_records
                        ]
                    )
            except (ValueError, IndexError) as e:
                logging.warning(f"Error processing deal:{deal_record}. Error:{e}")
    _write_results(output_path, csv_dicts, _get_headers(output_format))


if __name__ == "__main__":
    report()



import logging


def configure_logging(log_level: str):
    if log_level == "verbose":
        logging.basicConfig(level=logging.DEBUG)
    elif log_level == "quiet":
        logging.basicConfig(level=logging.CRITICAL)
    else:
        logging.basicConfig(level=logging.INFO)
